using System;
using System.Collections;
using System.Runtime.InteropServices;
using Mono.Unix;

namespace Banshee {
    static class Watcher 
    {
        private static Hashtable directoryMap;
        private static Hashtable fileMap;
        private static bool inExists = false;
        private enum Codes {
            Invalid, Changed, Deleted, StartExecuting, 
            StopExecuting, Created, Moved, Acknowledge,
            Exists, EndExists
        };

        [DllImport ("libwatch.so")]
        private static extern int init_watch();
    	
        [DllImport ("libwatch.so")]
        private static extern int destroy_watch();
    	
        [DllImport ("libwatch.so")]
        private static extern int watch_file(string pathname);
    	
        [DllImport ("libwatch.so")]
        private static extern int watch_directory(string pathname);

        [DllImport ("libwatch.so")]
        private static extern int stop_watch(int reqnum);
    	
        [DllImport ("libwatch.so")]
        private static extern int pending();
	
        [DllImport ("libwatch.so")]
        private unsafe static extern int next_event(int *reqnumptr, sbyte *filename, int *codeptr);
    	
        [DllImport ("libwatch.so")]
        private static extern int is_directory(string pathname);
        
        private static void Act(string fileName, int code)
        {
            System.Console.WriteLine(code + " happened to " + fileName);
        }
    	
        public static void InitWatch()
        {
            directoryMap = new Hashtable();
            fileMap = new Hashtable();
    		
            if(init_watch() != 0) {
                /*TODO*/
    		}
        }
    	
        public static void DestroyWatch()
        {
            foreach(int reqnum in directoryMap.Keys)
                stop_watch(reqnum);
    		
            foreach(int reqnum in fileMap.Keys)
                stop_watch(reqnum);

            destroy_watch();
    		
            directoryMap = null;
            fileMap = null;
        }

        public static void Watch(string pathname)
        {
            int reqnum;
            
            pathname = StringUtil.UriToFileName(pathname);
            bool isDirectory = (is_directory(pathname) == 1);
    		
            if(isDirectory) {
                reqnum = watch_directory(pathname);
                if (pathname[pathname.Length - 1] != '/')
                    pathname += '/';
            } else {
                reqnum = watch_file(pathname);
            }
    		
            if(reqnum < 0) {
                /*TODO*/
                return;
            }
    		
            if(isDirectory)
                directoryMap.Add(reqnum, pathname);
            else
                fileMap.Add(reqnum, pathname);
                
            System.Console.WriteLine("Watching " + pathname);
        }
    	
        public static void StopWatch(int reqnum)
        {
            if(stop_watch (reqnum) != 0) {
    	       /*TODO*/
            }
    		
            if(directoryMap.ContainsKey (reqnum))
                directoryMap.Remove (reqnum);
            else if(fileMap.ContainsKey (reqnum))
                fileMap.Remove (reqnum);
            else {
                /*TODO*/
            }
        }
        
        public static void StopWatch(string fileName)
        {
            bool found = false;
            int key = 0;
            
            fileName = StringUtil.UriToFileName(fileName);
            
            IDictionaryEnumerator dirEnum = directoryMap.GetEnumerator();
            IDictionaryEnumerator fileEnum = fileMap.GetEnumerator();
            
            while(dirEnum.MoveNext()) {
                if(dirEnum.Value as string == fileName) {
                    found = true;
                    key = (int) dirEnum.Key;
                    break;
                }
            }

            if(found) {
                StopWatch(key);
                return;
            }
            
            while(fileEnum.MoveNext()) {
                if(fileEnum.Value as string == fileName) {
                    found = true;
                    key = (int) fileEnum.Key;
                    break;
                }
            }
            
            if(found) {
                StopWatch(key);
                return;
            }

        }
    	
        private static int Pending()
        {           
            return pending();
        }
    	
        private static void NextEvent(out string fileName, out int code_)
        {
            int reqnum, code;
            int len;
            sbyte[] filename_bytes = new sbyte[4096];
    		
            unsafe {
                fixed(sbyte *filename = filename_bytes)
                {
                    next_event(&reqnum, filename, &code);
    				
                    for(len = 0; len < 4096 && filename_bytes[len] != '\0'; len++);
    		
                    fileName = new System.String(filename, 0, len);
                }
            }
    		
            if(directoryMap.ContainsKey(reqnum)) {
                code_ = code;
    			
                if(code_ == (int) Codes.Exists) {
                    if(inExists)
                        fileName = (directoryMap[reqnum] as string) + fileName;
                    else 
                        inExists = true;
                } else if(code_ == (int) Codes.EndExists) {
                    inExists = false;
                } else {
                    fileName = (directoryMap[reqnum] as string) + fileName;
                }
            } else if(fileMap.ContainsKey (reqnum)) {
    			code_ = code;
            } else {
                code_ = 0;
                /*TODO*/
            }
        }

        public static bool Update()
        {
            string fileName = "";
            int code = 0;
    		
            while(Watcher.Pending() > 0) {
                NextEvent(out fileName, out code);
                Act(fileName, code);
            }
            
            return true;
        }
    }
}
