#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fam.h>

#include "libwatch.h"

static FAMConnection fc;

static int request_num = 1;

int init_watch (void)
{
	return FAMOpen(&fc);
}

int destroy_watch (void)
{
	return FAMClose(&fc);
}

static int watch (const char *pathname, int (*monitor_func) (FAMConnection*, const char*, FAMRequest*))
{
	FAMRequest fr = {request_num};
		
	if (monitor_func (&fc, pathname, &fr) < 0)
		return -1;
		
	return request_num++;
}

int watch_file (char *pathname)
{
	return watch ((const char *) pathname, FAMMonitorFile2);
}

int watch_directory (char *pathname)
{
	return watch ((const char *) pathname, FAMMonitorDirectory2);
}

int stop_watch (int req)
{
	FAMRequest fr = {req};
	
	return FAMCancelMonitor (&fc, &fr);
}

int pending (void)
{
	return FAMPending (&fc);
}

int next_event (int *reqnumptr, char *filename, int *codeptr)
{
	FAMEvent fe;
	
	if (FAMNextEvent (&fc, &fe) < 0)
		return -1;
	
	*reqnumptr = fe.fr.reqnum;
	
	memcpy (filename, fe.filename, strlen (fe.filename) + 1);
	
	*codeptr = fe.code;
	
	return 0;
}

int is_directory (char *pathname)
{
	struct stat buf;
	
	if (stat (pathname, &buf) < 0)
		return -1;
		
	return !!S_ISDIR (buf.st_mode);
}

